﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Media;
using System.IO.Ports;

namespace PM_2408C2R
{
    public partial class Form_2408C2R : Form
    {
        const string STR_MODEL = " PM-2408C2R ";                // Model: PM-2408C2R

        /* Command ********************************************************** */
        const byte CMD_READ_DATA = 0x80;                        // Command: Read Data (讀取資料)
        const byte CMD_SYNC_EN   = 0x70;                        // Command: SYNC Enable (同步致能)
        const byte CMD_L1_SET    = 0x91;                        // Command: L1 Set (L1設定)
        const byte CMD_L2_SET    = 0x92;                        // Command: L2 Set (L2設定)
        const byte CMD_L1_OUTPUT = 0xA1;                        // Command: L1 Output (L1輸出電流)
        const byte CMD_L2_OUTPUT = 0xA2;                        // Command: L2 Output (L2輸出電流)
        const byte CMD_L1_ON_OFF = 0xC1;                        // Command: L1 ON/OFF
        const byte CMD_L2_ON_OFF = 0xC2;                        // Command: L2 ON/OFF

        /* ACK_State ======================================================== */
        const byte AS_FAILED  = 0x80;                           // bit7 命令執行狀態  : 0=成功, 1=失敗
        const byte AS_SYNC_EN = 0x20;                           // bit5 ON/OFF同步致能: 0=ASYNC (L1和L2 ON/OFF個別控制)
                                                                //                      1=SYNC  (由L1 ON/OFF同步控制)
        const byte AS_PWR_ERR = 0x10;                           // bit4 電源狀態      : 0=OK, 1=Error
        const byte AS_SYS_ERR = 0x01;                           // bit0 系統狀態      : 0=OK, 1=Error

        /* SYNC_EN ========================================================== */
        const byte ASYNC = 0;                                   // ASYNC (L1和L2 ON/OFF個別控制)
        const byte SYNC = 1;                                    // SYNC  (由L1 ON/OFF同步控制)

        /* Lx_Set =========================================================== */
        const byte SET_EXT_CTRL = 0x02;                         // bit1 外部ON/OFF控制: 0=Disable, 1=Enable
        const byte SET_ON_OFF   = 0x01;                         // bit0 ON/OFF        : 0=OFF, 1=ON

        /* ================================================================== */
        const ushort Lines_Max = 256;                           // 最大行數
        const byte ACK_TO = 10;                                 // 回應逾時, 約500ms (10 * 50ms)
        const byte AUTO_TMR = 60;                               // 自動詢問計時, 約3sec (60 * 50ms)

        /* ****************************************************************** */
        ushort Line_No = 0;                                     // 行數編號
        ushort Line_CTR = 0;                                    // 行數計數
        String[] PortNo;                                        // 埠編號
        bool PORT_Flag = false;                                 // PORT 旗標
        bool READ_Flag = false;                                 // READ 旗標
        byte ACK_TO_CTR = 0;                                    // 回應逾時計時
        byte ACK_State = 0x00;                                  // 回應狀態
        byte Auto_CTR = 0;                                      // 自動詢問計數
        byte Flash_CTR = 0;                                     // 閃爍計數
        int[] MaxCurrent = new int[] { 500, 600, 700, 800 };

        /* TX =============================================================== */
        bool TX_Flag = false;                                   // TX 旗標
        byte[] TX_Packet = new byte[7];                         // TX 封包
        byte TX_CMD = 0;                                        // TX 命令
        byte TX_CTR = 0;                                        // TX 傳送次數

        /* RX =============================================================== */
        bool RX_Flag = false;                                   // RX 旗標
        byte[] RX_Buffer = new byte[128];                       // RX 緩衝區
        byte[] RX_Packet = new byte[32];                        // RX 封包
        byte RX_Index = 0;                                      // RX 封包索引
        byte RX_Length = 0;                                     // RX 封包長度

        /* ================================================================== */
        bool SYNC_EN_CHG = false;                               // SYNC Enable 已改變
        byte SYNC_EN = ASYNC;                                   // SYNC Enable

        /* L1 =============================================================== */
        bool L1_Set_CHG = false;                                // L1 Set 已改變
        bool L1_Output_CHG = false;                             // L1 Output 已改變
        bool L1_ON_OFF_CHG = false;                             // L1 ON/OFF 已改變
        byte L1_Set = 0;                                        // L1 設定值
        int L1_Output = 0;                                      // L1 輸出值

        /* L2 =============================================================== */
        bool L2_Set_CHG = false;                                // L2 Set 已改變
        bool L2_Output_CHG = false;                             // L2 Output 已改變
        bool L2_ON_OFF_CHG = false;                             // L2 ON/OFF 已改變
        byte L2_Set = 0;                                        // L2 設定值
        int L2_Output = 0;                                      // L2 輸出值

        /* ****************************************************************** */

        public Form_2408C2R()
        {
            InitializeComponent();
        }

        private void Form_2408C2R_Load(object sender, EventArgs e)
        {
            cbb_L1_MaxC.SelectedIndex = 0;                      // 500mA
            cbb_L2_MaxC.SelectedIndex = 0;                      // 500mA
            COMPort_Check();                                    // 串列埠檢知
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            if (PORT_Flag == true)                              // 串列埠已開啟?
            {
                if (RX_Flag == true)                            // 已接收封包?
                {
                    TX_Flag = RX_Flag = false;
                    TX_CTR = ACK_TO_CTR = 0;
                    if (READ_Flag == false)
                    {
                        Update_Lx();                            // 更新 (Lx參數)
                        Update_ACK();                           // 更新 (ACK)
                        READ_Flag = true;
                    }
                    else Update_ACK();                          // 更新 (ACK)
                }
                else
                {
                    if (TX_Flag == true)                        // 已傳送命令?
                    {
                        Check_RX_TO();                          // 檢查接收逾時
                    }
                    else
                    {
                        Check_TX_CMD();                         // 檢查傳送命令
                    }
                    Auto_Ask();                                 // 自動詢問
                }
            }
            ERR_Flash();                                        // 錯誤提示閃爍
        }

        /* ================================================================== */

        private void Check_RX_TO()                              // 檢查接收逾時
        {
            if (++ACK_TO_CTR >= ACK_TO)
            {
                ACK_TO_CTR = 0;
                if (TX_CTR > 2)
                {
                    Dsp_String(Color.Red, STR_MODEL + "逾時未回應 !\n");
                    COMPort_Disconnect();                       // 串列埠離線
                    SystemSounds.Beep.Play();
                }
                else
                {
                    switch (TX_CMD)
                    {
                        case CMD_READ_DATA: TX_Read_Data(); break;  // 傳送: Read Data
                        case CMD_SYNC_EN  : TX_SYNC_EN(); break;    // 傳送: SYNC Enable
                        case CMD_L1_SET   : TX_L1_Set(); break;     // 傳送: L1 Set
                        case CMD_L2_SET   : TX_L2_Set(); break;     // 傳送: L2 Set
                        case CMD_L1_OUTPUT: TX_L1_Output(); break;  // 傳送: L1 Output
                        case CMD_L2_OUTPUT: TX_L2_Output(); break;  // 傳送: L2 Output
                        case CMD_L1_ON_OFF: TX_L1_ON_OFF(); break;  // 傳送: L1 ON/OFF
                        case CMD_L2_ON_OFF: TX_L2_ON_OFF(); break;  // 傳送: L2 ON/OFF
                    }
                }
            }
        }

        private void Check_TX_CMD()                             // 檢查傳送命令
        {
            if (SYNC_EN_CHG == true) { TX_SYNC_EN(); return; }
            if (L1_Set_CHG == true) { TX_L1_Set(); return; }
            if (L2_Set_CHG == true) { TX_L2_Set(); return; }
            if (L1_Output_CHG == true) { TX_L1_Output(); return; }
            if (L2_Output_CHG == true) { TX_L2_Output(); return; }
            if (L1_ON_OFF_CHG == true) { TX_L1_ON_OFF(); return; }
            if (L2_ON_OFF_CHG == true) { TX_L2_ON_OFF(); return; }
        }

        private void Auto_Ask()                                 // 自動詢問
        {
            if (++Auto_CTR >= AUTO_TMR) TX_Read_Data();         // 傳送: Read Data
            if (cbb_L1_MaxC.DroppedDown == true) Auto_CTR = 0;
            if (cbb_L2_MaxC.DroppedDown == true) Auto_CTR = 0;
            if (nud_L1_Output.Focused == true) Auto_CTR = 0;
            if (nud_L2_Output.Focused == true) Auto_CTR = 0;
        }

        private void ERR_Flash()                                // 錯誤提示閃爍
        {
            if (Flash_CTR == 0)
            {
                if ((ACK_State & AS_PWR_ERR) == AS_PWR_ERR) lb_PWR_ERR.BackColor = Color.Red;
                if ((ACK_State & AS_SYS_ERR) == AS_SYS_ERR) lb_L1.BackColor = lb_L2.BackColor = Color.Red;
            }
            if (Flash_CTR == 5)
            {
                if ((ACK_State & AS_PWR_ERR) == AS_PWR_ERR) lb_PWR_ERR.BackColor = Color.Black;
                if ((ACK_State & AS_SYS_ERR) == AS_SYS_ERR) lb_L1.BackColor = lb_L2.BackColor = Color.Black;
            }
            if (++Flash_CTR > 9) Flash_CTR = 0;
        }

        /* ================================================================== */

        private void Update_Lx()                                // 更新 (Lx參數)
        {
            Dsp_RX_Packet(SystemColors.WindowText, 11);

            if ((RX_Packet[3] & AS_SYNC_EN) == 0)
            {
                SYNC_EN = ASYNC;
                ckb_SYNC.Checked = false;
            }
            else
            {
                SYNC_EN = SYNC;
                ckb_SYNC.Checked =true;
            }

            Update_L1();                                        // 更新 (L1)
            Update_L2();                                        // 更新 (L2)
        }

        private void Update_L1()                                // 更新 (L1)
        {
            L1_Set = (byte)(RX_Packet[4] & 0x33);
            if ((L1_Set & SET_ON_OFF) == 0)                     // L1 ON/OFF ?
            {
                pn_L1_LED.BackColor = Color.Black;
                btn_L1.Text = "ON";
            }
            else
            {
                pn_L1_LED.BackColor = Color.Green;
                btn_L1.Text = "OFF";
            }
            if ((L1_Set & SET_EXT_CTRL) == 0)                   // L1外部ON/OFF控制 ?
            {
                ckb_L1_EXT.Checked = false;
                pn_L1_LED.Visible = btn_L1.Visible = true;
            }
            else
            {
                ckb_L1_EXT.Checked = true;
                pn_L1_LED.Visible = btn_L1.Visible = false;
            }

            cbb_L1_MaxC.SelectedIndex = (L1_Set >> 4) & 0x03;
            int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
            int Output = RX_Packet[5];
            Output <<= 8;
            Output |= RX_Packet[6];
            nud_L1_Output.Maximum = tkb_L1_Output.Maximum = MaxC;
            nud_L1_Output.Value = tkb_L1_Output.Value = Output;
            tkb_L1_Output.TickFrequency = MaxC / 10;
            lb_L1_PCV.Text = String.Format("{0:F1}", (float)(Output * 100) / MaxC);
            L1_Output = Output;
        }

        private void Update_L2()                                // 更新 (L2)
        {
            L2_Set = (byte)(RX_Packet[7] & 0x33);
            if ((L2_Set & SET_ON_OFF) == 0)                     // L2 ON/OFF ?
            {
                pn_L2_LED.BackColor = Color.Black;
                btn_L2.Text = "ON";
            }
            else
            {
                pn_L2_LED.BackColor = Color.Green;
                btn_L2.Text = "OFF";
            }
            if ((L2_Set & SET_EXT_CTRL) == 0)                   // L2外部ON/OFF控制 ?
            {
                ckb_L2_EXT.Checked = false;
                pn_L2_LED.Visible = btn_L2.Visible = true;
            }
            else
            {
                ckb_L2_EXT.Checked = true;
                pn_L2_LED.Visible = btn_L2.Visible = false;
            }

            cbb_L2_MaxC.SelectedIndex = (L2_Set >> 4) & 0x03;
            int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
            int Output = RX_Packet[8];
            Output <<= 8;
            Output |= RX_Packet[9];
            nud_L2_Output.Maximum = tkb_L2_Output.Maximum = MaxC;
            nud_L2_Output.Value = tkb_L2_Output.Value = Output;
            tkb_L2_Output.TickFrequency = MaxC / 10;
            lb_L2_PCV.Text = String.Format("{0:F1}", (float)(Output * 100) / MaxC);
            L2_Output = Output;
        }

        private void Update_ACK()                               // 更新 (ACK)
        {
            if (RX_Length == 5) Dsp_RX_Packet(SystemColors.WindowText, 5);

            ACK_State = RX_Packet[3];
            if ((ACK_State & 0x91) != 0) SystemSounds.Beep.Play();  // AS_FAILED | AS_PWR_ERR | AS_SYS_ERR

            if ((ACK_State & AS_FAILED) == AS_FAILED) Dsp_String(Color.Red, STR_MODEL + "命令執行失敗 !\n");

            if ((ACK_State & AS_SYS_ERR) == AS_SYS_ERR)
            {
                lb_L1.Text = lb_L2.Text = "System Error";
                Dsp_String(Color.Red, STR_MODEL + "系統錯誤 !\n");
            }
            else
            {
                lb_L1.Text = "L1";
                lb_L2.Text = "L2";
                lb_L1.BackColor = lb_L2.BackColor = Color.Black;
            }

            if ((ACK_State & AS_PWR_ERR) == AS_PWR_ERR)
            {
                lb_PWR_ERR.ForeColor = Color.White;
                Dsp_String(Color.Red, STR_MODEL + "電源過低 !\n");
            }
            else
            {
                lb_PWR_ERR.ForeColor = Color.Gray;
                lb_PWR_ERR.BackColor = Color.Black;
            }

            if ((ACK_State & 0x11) != 0)                        // AS_PWR_ERR | AS_SYS_ERR
            {
                gpb_ON_OFF.Enabled = gpb_L1.Enabled = gpb_L2.Enabled = false;
            }
            else
            {
                gpb_ON_OFF.Enabled = gpb_L1.Enabled = gpb_L2.Enabled = true;
            }
        }

        /* RichTextBox 顯示 ************************************************* */

        private void Dsp_String(Color text_color, string str)   // 顯示字串
        {
            if (++Line_CTR > Lines_Max) { Line_CTR = 0; richTextBox.Clear(); }

            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();
            richTextBox.SelectionBackColor = Color.Gainsboro;
            richTextBox.SelectionColor = SystemColors.WindowText;
            richTextBox.AppendText(String.Format("{0:X4}", Line_No++));
            richTextBox.SelectionBackColor = SystemColors.Window;
            richTextBox.SelectionColor = text_color;
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        private void Dsp_TX_Packet(Color text_color, byte length, byte counter)  // 顯示 TX 封包
        {
            if (++Line_CTR > Lines_Max) { Line_CTR = 0; richTextBox.Clear(); }

            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();
            richTextBox.SelectionBackColor = Color.Gainsboro;
            richTextBox.SelectionColor = SystemColors.WindowText;
            richTextBox.AppendText(String.Format("{0:X4}", Line_No++));
            richTextBox.SelectionBackColor = SystemColors.Window;
            richTextBox.SelectionColor = text_color;

            string str = " TX:";
            for (byte i = 0; i < length; i++) str += String.Format(" {0:X2}", TX_Packet[i]);
            str += " (" + Convert.ToString(counter) + ")\n";
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        private void Dsp_RX_Packet(Color text_color, byte length)  // 顯示 RX 封包
        {
            if (++Line_CTR > Lines_Max) { Line_CTR = 0; richTextBox.Clear(); }

            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();
            richTextBox.SelectionBackColor = Color.Gainsboro;
            richTextBox.SelectionColor = SystemColors.WindowText;
            richTextBox.AppendText(String.Format("{0:X4}", Line_No++));
            richTextBox.SelectionBackColor = SystemColors.Window;
            richTextBox.SelectionColor = text_color;

            string str = " RX:";
            for (byte i = 0; i < length; i++) str += String.Format(" {0:X2}", RX_Packet[i]);
            str += "\n";
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        /* 串列埠 *********************************************************** */

        private void cbb_PortNo_Click(object sender, EventArgs e)
        {
            COMPort_Check();                                    // 串列埠檢知
        }

        private void btn_Connect_Click(object sender, EventArgs e)
        {
            bool state = false;

            if (serialPort.IsOpen)                              // 串列埠是否已開啟?
            {
                COMPort_Disconnect();                           // 串列埠離線
            }
            else
            {
                PortNo = SerialPort.GetPortNames();             // 取得存在的 COM Port.
                if (PortNo.Length >= 1)
                {
                    if (cbb_PortNo.Text == "")
                    {
                        Dsp_String(Color.Red, " 未選擇串列埠 !\n");
                        SystemSounds.Beep.Play();
                    }
                    else
                    {
                        foreach (string port in PortNo)
                        {
                            if (port == cbb_PortNo.Text) { state = true; break; }
                        }
                        if (state == true)
                        {
                            if (serialPort.IsOpen)
                            {
                                COMPort_Disconnect();           // 串列埠離線
                            }
                            else
                            {
                                COMPort_Connect();              // 串列埠連線
                                if (PORT_Flag == true)
                                {
                                    TX_Read_Data();             // 傳送: Read Data
                                }
                            }
                        }
                    }
                }
                else
                {
                    cbb_PortNo.Items.Clear();                   // 清除 ComboBox.Items 的內容.
                    Dsp_String(Color.Red, " 未發現任何串列埠 !\n");
                    SystemSounds.Beep.Play();
                }
            }
        }

        private void COMPort_Check()                            // 串列埠檢知
        {
            string Port_No = cbb_PortNo.Text;

            PortNo = SerialPort.GetPortNames();                 // 取得存在的 COM Port.
            cbb_PortNo.Items.Clear();                           // 清除 ComboBox.Items 的內容.
            if (PortNo.Length >= 1)
            {
                foreach (string port in PortNo)                 // 將找到的現有 COM Port 加入 ComboBox.Items
                {
                    cbb_PortNo.Items.Add(port);
                    if (port == Port_No) cbb_PortNo.Text = Port_No;
                }
                if (cbb_PortNo.Text == "")
                {
                    cbb_PortNo.Text = PortNo[0];                // ComboBox.Text 先顯示一個現存的 COM Port.
                }
            }
            else
            {
                Dsp_String(Color.Red, " 未發現任何串列埠 !\n");
                SystemSounds.Beep.Play();
            }
        }

        private void COMPort_Connect()                          // 串列埠連線
        {
            TX_Flag = RX_Flag = PORT_Flag = false;
            TX_CMD = TX_CTR = RX_Index = 0;
            SYNC_EN_CHG = false;
            L1_Set_CHG = L1_Output_CHG = L1_ON_OFF_CHG = false;
            L2_Set_CHG = L2_Output_CHG = L2_ON_OFF_CHG = false;
            serialPort.PortName = cbb_PortNo.Text;

            /* 設定串列埠鮑率 ----------------------------------------------- */
            if (rbtn_BR115200.Checked == true) serialPort.BaudRate = 115200;
            else serialPort.BaudRate = 9600;

            try
            {
                serialPort.Open();                              // 開啟新的序列埠連線.
                serialPort.DiscardInBuffer();                   // 捨棄序列驅動程式接收緩衝區的資料.
                serialPort.DiscardOutBuffer();                  // 捨棄序列驅動程式傳輸緩衝區的資料.
                cbb_PortNo.Enabled = rbtn_BR9600.Enabled = rbtn_BR115200.Enabled = false;
                btn_Connect.Text = "Disconnect";
                tss_Status.Text = "Connect";
                Dsp_String(Color.Magenta, " " + cbb_PortNo.Text + " connect.\n");
                PORT_Flag = true;
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 開啟串列埠失敗 !\n");
                SystemSounds.Beep.Play();
            }
        }

        private void COMPort_Disconnect()                       // 串列埠離線
        {
            TX_Flag = RX_Flag = PORT_Flag = false;
            TX_CMD = TX_CTR = RX_Index = ACK_State = 0;
            SYNC_EN_CHG = false;
            L1_Set_CHG = L1_Output_CHG = L1_ON_OFF_CHG = false;
            L2_Set_CHG = L2_Output_CHG = L2_ON_OFF_CHG = false;

            try
            {
                serialPort.Close();                             // 關閉連接埠連線.
                Dsp_String(Color.Magenta, " " + cbb_PortNo.Text + " disconnect.\n");
            }
            catch (Exception)
            {
                COMPort_Check();                                // 串列埠檢知
            }

            cbb_PortNo.Enabled = rbtn_BR9600.Enabled = rbtn_BR115200.Enabled = true;
            btn_Connect.Text = "Connect";
            tss_Status.Text = "Disconnect";
            lb_L1.Text = "L1";
            lb_L2.Text = "L2";
            lb_L1.BackColor = lb_L2.BackColor = Color.Black;
            lb_PWR_ERR.ForeColor = Color.Gray;
            lb_PWR_ERR.BackColor = Color.Black;
            gpb_ON_OFF.Enabled = gpb_L1.Enabled = gpb_L2.Enabled = false;
        }

        /* 串列埠傳送 ******************************************************* */

        private void TX_Read_Data()                             // 傳送: Read Data
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 5;                               // Packet Length
                TX_Packet[3] = CMD_READ_DATA;                   // Command
                for (i = sum = 0; i < 4; i++) sum += TX_Packet[i];
                TX_Packet[4] = sum;                             // Checksum
                TX_CMD = CMD_READ_DATA;
                TX_CTR = 0;
                TX_Flag = true;
                READ_Flag = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Black, 5, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 5, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 5);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (Read Data), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_SYNC_EN()                               // 傳送: SYNC Enable
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // Packet Length
                TX_Packet[3] = CMD_SYNC_EN;                     // Command
                TX_Packet[4] = SYNC_EN;
                for (i = sum = 0; i < 5; i++) sum += TX_Packet[i];
                TX_Packet[5] = sum;                             // Checksum
                TX_CMD = CMD_SYNC_EN;
                TX_CTR = 0;
                TX_Flag = true;
                SYNC_EN_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Black, 6, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 6, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (SYNC Enable), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L1_Set()                                // 傳送: L1 Set
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // Packet Length
                TX_Packet[3] = CMD_L1_SET;                      // Command
                TX_Packet[4] = L1_Set;                          // L1 設定值
                for (i = sum = 0; i < 5; i++) sum += TX_Packet[i];
                TX_Packet[5] = sum;                             // Checksum
                TX_CMD = CMD_L1_SET;
                TX_CTR = 0;
                TX_Flag = true;
                L1_Set_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Maroon, 6, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 6, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L1 Set), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L2_Set()                                // 傳送: L2 Set
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // Packet Length
                TX_Packet[3] = CMD_L2_SET;                      // Command
                TX_Packet[4] = L2_Set;                          // L2 設定值
                for (i = sum = 0; i < 5; i++) sum += TX_Packet[i];
                TX_Packet[5] = sum;                             // Checksum
                TX_CMD = CMD_L2_SET;
                TX_CTR = 0;
                TX_Flag = true;
                L2_Set_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Maroon, 6, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 6, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L2 Set), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L1_Output()                             // 傳送: L1 Output
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 7;                               // Packet Length
                TX_Packet[3] = CMD_L1_OUTPUT;                   // Command
                TX_Packet[4] = (byte)(L1_Output >> 8);          // L1 Output, HB
                TX_Packet[5] = (byte)L1_Output;                 // L1 Output, LB
                for (i = sum = 0; i < 6; i++) sum += TX_Packet[i];
                TX_Packet[6] = sum;                             // Checksum
                TX_CMD = CMD_L1_OUTPUT;
                TX_CTR = 0;
                TX_Flag = true;
                L1_Output_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Blue, 7, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 7, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 7);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L1 Output), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L2_Output()                             // 傳送: L2 Output
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 7;                               // Packet Length
                TX_Packet[3] = CMD_L2_OUTPUT;                   // Command
                TX_Packet[4] = (byte)(L2_Output >> 8);          // L2 Output, HB
                TX_Packet[5] = (byte)L2_Output;                 // L2 Output, LB
                for (i = sum = 0; i < 6; i++) sum += TX_Packet[i];
                TX_Packet[6] = sum;                             // Checksum
                TX_CMD = CMD_L2_OUTPUT;
                TX_CTR = 0;
                TX_Flag = true;
                L2_Output_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Blue, 7, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 7, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 7);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L2 Output), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L1_ON_OFF()                             // 傳送: L1 ON/OFF
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // Packet Length
                TX_Packet[3] = CMD_L1_ON_OFF;                   // Command
                TX_Packet[4] = (byte)(L1_Set & SET_ON_OFF);     // L1 ON/OFF
                for (i = sum = 0; i < 5; i++) sum += TX_Packet[i];
                TX_Packet[5] = sum;                             // Checksum
                TX_CMD = CMD_L1_ON_OFF;
                TX_CTR = 0;
                TX_Flag = true;
                L1_ON_OFF_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Green, 6, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 6, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L1 ON/OFF), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        private void TX_L2_ON_OFF()                             // 傳送: L2 ON/OFF
        {
            byte i, sum;

            if (TX_Flag == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // Packet Length
                TX_Packet[3] = CMD_L2_ON_OFF;                   // Command
                TX_Packet[4] = (byte)(L2_Set & SET_ON_OFF);     // L2 ON/OFF
                for (i = sum = 0; i < 5; i++) sum += TX_Packet[i];
                TX_Packet[5] = sum;                             // Checksum
                TX_CMD = CMD_L2_ON_OFF;
                TX_CTR = 0;
                TX_Flag = true;
                L2_ON_OFF_CHG = false;
            }

            if (TX_CTR == 0) Dsp_TX_Packet(Color.Green, 6, ++TX_CTR);
            else Dsp_TX_Packet(Color.Magenta, 6, ++TX_CTR);
            ACK_TO_CTR = Auto_CTR = 0;

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                Dsp_String(Color.Red, " 傳送封包 (L2 ON/OFF), 發生不可預期的錯誤 !\n");
                SystemSounds.Beep.Play();
                COMPort_Disconnect();                           // 串列埠離線
            }
        }

        /* 串列埠接收 ******************************************************* */

        private void serialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            int bytes = serialPort.BytesToRead;

            if (bytes > 0)
            {
                serialPort.Read(RX_Buffer, 0, bytes);
                Packet_Check(bytes);                            // 封包資料判斷
            }
        }

        private void Packet_Check(int length)                   // 封包資料判斷
        {
            byte i, n, sum;

            for (i = 0; i < length; i++)
            {
                RX_Packet[RX_Index] = RX_Buffer[i];
                if (RX_Index == 0)                              // Header 'S'
                {
                    if (RX_Packet[0] == 0x53) RX_Index++;
                }
                else
                {
                    if (RX_Index == 1)                          // Header 'M'
                    {
                        if (RX_Packet[1] == 0x4D) RX_Index++;
                        else RX_Index = 0;
                    }
                    else
                    {
                        if (RX_Index == 2)                      // Packet Length
                        {
                            if ((RX_Packet[2] == 11) || (RX_Packet[2] == 5))
                            {
                                RX_Index++;
                                RX_Length = RX_Packet[2];
                            }
                            else RX_Index = 0;
                        }
                        else
                        {
                            if (++RX_Index >= RX_Length)
                            {
                                RX_Index = 0;
                                for (n = sum = 0; n < (RX_Length - 1); n++) sum += RX_Packet[n];
                                if (sum == RX_Packet[n]) Packet_OK();  // 封包資料格式 OK
                            }
                        }
                    }
                }
            }
        }

        private void Packet_OK()                                // 封包資料格式 OK
        {
            switch (TX_CMD)
            {
                case CMD_READ_DATA: if (RX_Length == 11) RX_Flag = true;
                                    break;
                case CMD_SYNC_EN  : if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L1_SET   : if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L2_SET   : if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L1_OUTPUT: if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L2_OUTPUT: if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L1_ON_OFF: if (RX_Length == 5) RX_Flag = true;
                                    break;
                case CMD_L2_ON_OFF: if (RX_Length == 5) RX_Flag = true;
                                    break;
            }
        }

        /* L1輸出調整 ******************************************************* */

        private void cbb_L1_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L1_Set >> 4) & 0x03) != cbb_L1_MaxC.SelectedIndex)
            {
                byte v = (byte)(L1_Set & (SET_EXT_CTRL | SET_ON_OFF));
                L1_Set = (byte)((cbb_L1_MaxC.SelectedIndex << 4) | v);

                nud_L1_Output.Value = tkb_L1_Output.Value = L1_Output = 0;
                nud_L1_Output.Maximum = tkb_L1_Output.Maximum = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
                tkb_L1_Output.TickFrequency = tkb_L1_Output.Maximum / 10;
                L1_Output_CHG = false;

                if (READ_Flag == true) L1_Set_CHG = true;
            }
        }

        private void nud_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L1_Output.Value = (int)nud_L1_Output.Value;
            L1_Output = tkb_L1_Output.Value;
            int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
            lb_L1_PCV.Text = String.Format("{0:F1}", (float)(L1_Output * 100) / MaxC);

            if (READ_Flag == true) L1_Output_CHG = true;
        }

        private void tkb_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            nud_L1_Output.Value = tkb_L1_Output.Value;
            L1_Output = tkb_L1_Output.Value;
            int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
            lb_L1_PCV.Text = String.Format("{0:F1}", (float)(L1_Output * 100) / MaxC);

            if (READ_Flag == true) L1_Output_CHG = true;
        }

        /* L1 ON/OFF ======================================================== */

        private void ckb_L1_EXT_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L1_EXT.Checked == false)                    // Enable to Disable
            {
                pn_L1_LED.Visible = btn_L1.Visible = true;
                L1_Set &= (SET_EXT_CTRL ^ 0xFF);                // L1外部ON/OFF控制 = Disable
            }
            else                                                // Disable to Enable
            {
                pn_L1_LED.Visible = btn_L1.Visible = false;
                L1_Set |= SET_EXT_CTRL;                         // L1外部ON/OFF控制 = Enable
            }

            if (READ_Flag == true) L1_Set_CHG = true;

            if (SYNC_EN == SYNC)
            {
                ckb_L2_EXT.Checked = ckb_L1_EXT.Checked;
            }
        }

        private void btn_L1_Click(object sender, EventArgs e)
        {
            if ((L1_Set & SET_ON_OFF) == 0)                     // OFF to ON
            {
                pn_L1_LED.BackColor = Color.Green;
                btn_L1.Text = "OFF";
                L1_Set |= SET_ON_OFF;                           // L1 ON/OFF = ON
            }
            else                                                // ON to OFF
            {
                pn_L1_LED.BackColor = Color.Black;
                btn_L1.Text = "ON";
                L1_Set &= (SET_ON_OFF ^ 0xFF);                  // L1 ON/OFF = OFF
            }

            if (READ_Flag == true) L1_ON_OFF_CHG = true;

            if (SYNC_EN == SYNC)
            {
                pn_L2_LED.BackColor = pn_L1_LED.BackColor;
                btn_L2.Text = btn_L1.Text;
                if ((L1_Set & SET_ON_OFF) == 0) L2_Set &= (SET_ON_OFF ^ 0xFF);
                else L2_Set |= SET_ON_OFF;
            }
        }

        /* L2輸出調整 ******************************************************* */

        private void cbb_L2_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L2_Set >> 4) & 0x03) != cbb_L2_MaxC.SelectedIndex)
            {
                byte v = (byte)(L2_Set & (SET_EXT_CTRL | SET_ON_OFF));
                L2_Set = (byte)((cbb_L2_MaxC.SelectedIndex << 4) | v);

                nud_L2_Output.Value = tkb_L2_Output.Value = L2_Output = 0;
                nud_L2_Output.Maximum = tkb_L2_Output.Maximum = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
                tkb_L2_Output.TickFrequency = tkb_L2_Output.Maximum / 10;
                L2_Output_CHG = false;

                if (READ_Flag == true) L2_Set_CHG = true;
            }
        }

        private void nud_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L2_Output.Value = (int)nud_L2_Output.Value;
            L2_Output = tkb_L2_Output.Value;
            int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
            lb_L2_PCV.Text = String.Format("{0:F1}", (float)(L2_Output * 100) / MaxC);

            if (READ_Flag == true) L2_Output_CHG = true;
        }

        private void tkb_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            nud_L2_Output.Value = tkb_L2_Output.Value;
            L2_Output = tkb_L2_Output.Value;
            int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
            lb_L2_PCV.Text = String.Format("{0:F1}", (float)(L2_Output * 100) / MaxC);

            if (READ_Flag == true) L2_Output_CHG = true;
        }

        /* L2 ON/OFF ======================================================== */

        private void ckb_L2_EXT_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L2_EXT.Checked == false)                    // Enable to Disable
            {
                pn_L2_LED.Visible = btn_L2.Visible = true;
                L2_Set &= (SET_EXT_CTRL ^ 0xFF);                // L2外部ON/OFF控制 = Disable
            }
            else                                                // Disable to Enable
            {
                pn_L2_LED.Visible = btn_L2.Visible = false;
                L2_Set |= SET_EXT_CTRL;                         // L2外部ON/OFF控制 = Enable
            }

            if ((READ_Flag == true) && (SYNC_EN == ASYNC)) L2_Set_CHG = true;
        }

        private void btn_L2_Click(object sender, EventArgs e)
        {
            if ((L2_Set & SET_ON_OFF) == 0)                     // OFF to ON
            {
                pn_L2_LED.BackColor = Color.Green;
                btn_L2.Text = "OFF";
                L2_Set |= SET_ON_OFF;                           // L2 ON/OFF = ON
            }
            else                                                // ON to OFF
            {
                pn_L2_LED.BackColor = Color.Black;
                btn_L2.Text = "ON";
                L2_Set &= (SET_ON_OFF ^ 0xFF);                  // L2 ON/OFF = OFF
            }

            if ((READ_Flag == true) && (SYNC_EN == ASYNC)) L2_ON_OFF_CHG = true;
        }

        /* SYNC Enable ****************************************************** */

        private void ckb_SYNC_CheckedChanged(object sender, EventArgs e)
        {
            SystemSounds.Exclamation.Play();
            if (ckb_SYNC.Checked == false)
            {
                SYNC_EN = ASYNC;
                ckb_L2_EXT.Enabled = btn_L2.Enabled = true;
            }
            else
            {
                SYNC_EN = SYNC;
                ckb_L2_EXT.Enabled = btn_L2.Enabled = false;

                ckb_L2_EXT.Checked = ckb_L1_EXT.Checked;
                pn_L2_LED.BackColor = pn_L1_LED.BackColor;
                btn_L2.Text = btn_L1.Text;
                btn_L2.Visible = btn_L1.Visible;
                if ((L1_Set & SET_EXT_CTRL) == 0) L2_Set &= (SET_EXT_CTRL ^ 0xFF);
                else L2_Set |= SET_EXT_CTRL;
                if ((L1_Set & SET_ON_OFF) == 0) L2_Set &= (SET_ON_OFF ^ 0xFF);
                else L2_Set |= SET_ON_OFF;
            }

            if (READ_Flag == true) SYNC_EN_CHG = true;
        }
    }
}
